# -*- coding: utf-8 -*-
"""
Created on 2022/1/18
@project: SPAIC
@filename: Topology
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
from ..Network.BaseModule import BaseModule
from ..Network.Assembly import Assembly
# from ..Network.Synapse import Flatten_synapse
from ..Network.BaseModule import VariableAgent, Op
from ..Backend.Backend import Backend
from abc import abstractmethod
from typing import List
import numpy as np
from abc import ABC


class Projection(BaseModule):
    '''
    Class for projection between assemblies, which contain multiple connections between sub neurongroups
    '''
    _class_label = '<prj>'
    _con_count = 0

    def __init__(self, pre: Assembly, post: Assembly, policies=[], link_type=None, ConnectionParameters=None,
                 name=None):
        super(Projection, self).__init__()
        assert isinstance(pre, Assembly)
        assert isinstance(post, Assembly)
        self.name = name
        self.super = None
        self.abstract_level = 0
        self._backend: Backend = None
        self.pre = pre
        self.post = post
        self._connections = dict()
        self._projections = dict()

        if (pre.id is not None) and (post.id is not None):
            self.key = pre.id + "->" + post.id
        else:
            self.key = 'default' + str(Projection._con_count)
        Projection._con_count += 1
        self._leaf_connections = dict()

        if isinstance(ConnectionParameters, dict):
            self.ConnectionParameters = ConnectionParameters
        else:
            self.ConnectionParameters = dict()
        if pre._is_terminal and post._is_terminal:
            self.is_unit = True
        else:
            self.is_unit = False

        self._policies = policies
        if self._policies:
            assert link_type is not None
        if isinstance(self._policies, ConnectPolicy):
            self._policies = [self._policies]
        self.link_type = link_type
        self._supers = []

    def __setattr__(self, name, value):
        super(Projection, self).__setattr__(name, value)
        if name == 'super':
            return
        if isinstance(value, Connection):
            if self._backend: self._backend.builded = False
            self._connections[name] = value
            # self._leaf_connections[value.key] = value
            value.set_name(name)
            value.add_super(self)
        elif isinstance(value, Projection):
            if self._backend: self._backend.builded = False
            self._projections[name] = value
            # self._leaf_connections[value.key] = value
            value.set_name(name)
            value.add_super(self)

    def homologous(self, other):
        if (other.pre is self.pre) and (other.post is self.post):
            return True
        else:
            return False

    def is_empty(self):
        if self.is_unit:
            return False
        elif len(self._connections) + len(self._projections) + len(self._leaf_connections) == 0:
            return True
        else:
            return False

    def __and__(self, other):
        if not self.homologous(other):
            raise ValueError("can't do & operation for nonhomologous connections")
        if self.is_empty():
            new_connect = self
        elif other.is_empty():
            new_connect = other
        else:
            new_connect = Projection(self.pre, self.post)
            # unit connections
            key1 = set(self._connections.keys())
            key2 = set(other._connections.keys())
            unite_keys = key1.intersection(key2)
            for key in unite_keys:
                new_connect._connections[key] = self._connections[key]
                new_connect._connections[key].super = new_connect

            # assb connections
            key1 = set(self._projections.keys())
            key2 = set(other._projections.keys())
            unite_keys = key1.intersection(key2)
            for key in unite_keys:
                new_connect._projections[key] = self._projections[key] & other._projections[key]
                new_connect._projections[key].super = new_connect

            # leaf_connections
            key1 = set(self._leaf_connections.keys())
            key2 = set(other._leaf_connections.keys())
            unite_keys = key1.intersection(key2)
            for key in unite_keys:
                new_connect._leaf_connections[key] = self._leaf_connections[key] & other._leaf_connections[key]

        return new_connect

    def add_connection(self, con, name=None):
        assert isinstance(con, Projection)
        con.super = self
        key = con.key
        if not (con.pre in self.pre and con.post in self.post):
            raise ValueError(
                "the sub connection is not belong to this connection group (pre and post is not a member of the connected Assemblies)")

        con.set_name(name)
        if con.is_unit:
            self.__setattr__(name, con)
            # self._connections[key] = con
        else:
            self._projections[key] = con

    def add_leaf_connection(self, con):
        assert isinstance(con, Projection)
        con.super = self
        key = con.key
        if not (con.pre in self.pre and con.post in self.post):
            raise ValueError("the sub connection is not belong to this connection group (pre and post is not a member of the connected Assemblies)")
        con.set_name(None)
        self._leaf_connections[key] = con

    def del_connection(self):
        if self.super is not None:
            if self.key in self.super._connections:
                self.super._connections.pop(self.key)
            if self.key in self.super._projections:
                self.super._projections.pop(self.key)
            if self.key in self.super._leaf_connections:
                self.super._leaf_connections.pop(self.key)

            top_leaf = self.top._leaf_connections
            if self.key in top_leaf:
                top_leaf.pop(self.key)
            if self.super.is_empty():
                self.super.del_connection()
            self.super = None
        else:
            self.is_unit = False

    def expand_connection(self, to_level=-1):
        assert isinstance(to_level, int), ValueError("level is not int")
        if not bool(self._leaf_connections):
            self._leaf_connections[self.key] = self
        if self.is_unit:
            return self._leaf_connections.values()
        else:
            new_leaf_connections = dict()
            assb_connections = self._leaf_connections.values()
            while (len(assb_connections) > 0):
                if to_level >= 0 and to_level <= self.abstract_level:
                    break
                self.abstract_level += 1
                new_assb_connections = []
                for con in assb_connections:
                    if con.is_unit:
                        new_leaf_connections[con.key] = con
                    else:
                        pre_groups = con.pre.get_groups(recursive=False)
                        post_groups = con.post.get_groups(recursive=False)
                        for pre in pre_groups:
                            for post in post_groups:
                                new_con = Projection(pre, post)
                                # con.add_connection(new_con)
                                if new_con.is_unit:
                                    new_leaf_connections[new_con.key] = new_con
                                else:
                                    new_assb_connections.append(new_con)

                assb_connections = new_assb_connections

            for con in assb_connections:
                # 对于超过abstract_level的connection，都认为是leaf_connection
                new_leaf_connections[con.key] = con

            # self._leaf_connections = new_leaf_connections
            self._leaf_connections.clear()
            for key, con in new_leaf_connections.items():
                self.add_leaf_connection(con)

            return list(new_leaf_connections.values())

    def assembly_linked(self, assembly):
        if (assembly is self.pre) or (assembly is self.post):
            return True
        else:
            return False

    def replace_assembly(self, old_assembly, new_assembly):
        if old_assembly is self.pre:
            self.pre = new_assembly
        elif old_assembly is self.post:
            self.post = new_assembly
        else:
            raise ValueError("the old_assembly is not in the connnection")

    def add_super(self, assembly):
        assert isinstance(assembly, Assembly) or isinstance(assembly,
                                                            Projection), "the super is not Assembly or Projection"
        self._supers.append(assembly)

    def del_super(self, assembly):
        assert assembly in self._supers, "the assembly is not in supers"
        self._supers.remove(assembly)

    def get_connections(self, recursive=True):
        """
            Get the Connections in this assembly
        Args:
            recursive(bool): flag to decide if member connections of the member assemblies should be returned.

        Returns:
            List of Connections
        """
        if not recursive:
            return self._connections.values()
        else:
            connections = list(self._connections.values())
            for proj in self.sub_projections:
                connections.extend(proj.get_connections(recursive=False))
            return connections

    @property
    def top(self):
        if self.super is None:
            return self
        else:
            return self.super.top

    @property
    def sub_connections(self):
        if self.is_unit:
            raise ValueError("no sub_connections for unit connections")
        else:
            sub_connections = []
            sub_connections.extend(self._connections.values())
            return sub_connections

    @property
    def sub_projections(self):
        if self.is_unit:
            raise ValueError("no sub assb_connections for unit connections")
        else:
            return list(self._projections.values())

    @property
    def all_connections(self):
        if self.is_unit:
            return [self]
        else:
            unit_connections = []
            unit_connections.extend(self._connections.values())
            for con in self._projections.values():
                unit_connections.extend(con.all_connections)
            return unit_connections

    @property
    def all_projections(self):
        if self.is_unit:
            raise ValueError("no sub_connections for unit connections")
        else:
            assb_connections = []
            assb_connections.extend(self._projections.values())
            for con in self._projections.values():
                assb_connections.extend(con.all_projections)
            return assb_connections

    @property
    def leaf_connections(self):
        return list(self._leaf_connections.values())

    def get_str(self, level):

        # TODO: add str of subconnections
        level_space = "" + '-' * level
        repr_str = level_space + "|name:{}, type:{}, ".format(self.name, type(self).__name__)
        repr_str += "pre:{}, ".format(self.pre.name)
        repr_str += "post:{}\n ".format(self.post.name)
        level += 1
        for c in self._connections.values():
            repr_str += c.get_str(level)
        return repr_str

    def train(self, mode=True):
        self.training = mode
        for p in self._projections.values():
            p.train(mode)
        for c in self._connections.values():
            c.training = mode

    def build(self, backend=None):
        self._backend = backend

        # clear previous builded connections
        if backend.builded == True:
            for super_asb in self._supers:
                for con in self._connections:
                    super_asb.del_connection(con)
        if self._connections and not self._policies:
            return
        connection_inforamtion = self
        connection_inforamtion.expand_connection()
        for p in self._policies:
            connection_inforamtion = connection_inforamtion & p.generate_connection(self.pre,
                                                                                    self.post)

        self._connections = connection_inforamtion._connections
        self._projections = connection_inforamtion._projections
        self._leaf_connections = connection_inforamtion._leaf_connections
        all_connections = dict()
        all_connections.update(self._connections)
        all_connections.update(self._leaf_connections)
        new_connections = dict()
        auto_con_num = 0
        for key, con in all_connections.items():
            assert con.is_unit
            if con.link_type is not None:
                link_type = con.link_type
                con_name = con.name
                con_kwargs = con.con_kwargs
                con_kwargs.update(self.ConnectionParameters)
                syn_kwargs = con.syn_kwargs
            else:
                assert self.link_type is not None
                link_type = self.link_type
                con_name = 'auto_con' + str(auto_con_num)
                auto_con_num += 1
                con_kwargs = self.ConnectionParameters
                syn_kwargs = None

            new_connections[key] = Connection(pre=con.pre, post=con.post,
                                              name=con_name, link_type=link_type, syn_kwargs=syn_kwargs, **con_kwargs)
            new_connections[key].add_super(self)
        self._connections = new_connections



class SynapseModel(ABC):
    '''
    op -> (return_name, operation_name, input_name1, input_name2...)
    '''

    #: A dictionary mapping synapse model names to `Model` objects
    synapse_models = dict()

    def __init__(self, conn, **kwargs):
        super(SynapseModel, self).__init__()
        self.name = 'none'
        self._syn_operations = []
        self._syn_variables = dict()
        self._syn_constant_variables = dict()
        self._syn_tau_variables = dict()
        if conn is not None:
            updated_input = conn.updated_input
            if updated_input:
                self.input_name = conn.pre_var_name + '[input][updated]'
            else:
                self.input_name = conn.pre_var_name + '[input]'

    @staticmethod
    def register(name, model):
        '''
        Register a synapse model. Registered synapse models can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. `'basic'`)
        model : `SynapseModel`
            The synapse model object, e.g. an `Basic_synapse`.
        '''

        # only deal with lower case names -- we don't want to have 'LIF' and
        # 'lif', for example
        name = name.lower()
        if name in SynapseModel.synapse_models:
            raise ValueError(('A synapse_model with the name "%s" has already been registered') % name)

        if not issubclass(model, SynapseModel):
            raise ValueError(('Given model of type %s does not seem to be a valid SynapseModel.' % str(type(model))))

        SynapseModel.synapse_models[name] = model
        model.name = name

    @staticmethod
    def apply_model(model_name):
        '''
        Parameters
        ----------
        model_name : str
        Returns
        -------
        '''
        model_name = model_name.lower()
        if model_name not in SynapseModel.synapse_models:
            raise ValueError(('Given synapse model name is not in the model list'))
        else:
            return SynapseModel.synapse_models[model_name]


class Connection(Projection):
    '''
    Base class for all kinds of connections, including full connection, sparse connection, conv connection,....
    Ten connection methods are provided, as shown below (key: class):
        'full', FullConnection
        'one_to_one_sparse', one_to_one_sparse
        'one_to_one', one_to_one_mask
        'conv', conv_connect
        'sparse_connection_sparse', sparse_connect_sparse
        'sparse_connection', sparse_connect_mask
        'random_connection_sparse', random_connect_sparse
        'random_connection', random_connect_mask
    Args:
        pre(Assembly): the assembly which needs to be connected.
        post(Assembly): the assembly which needs to connect the pre.
        link_type(str): the type for connection: full, sparse, conv...

    Attributes:
        pre_group(groups): the neuron group which need to be connected in the pre.
        post_group(groups): the neuron group which need to connect with pre_group neuron.
        _var_names(list): a list contain variable names.

    Methods:
        __new__: before build a new connection, do some checks.
        get_var_names: get variable names.
        register: register a connection class.
        build: add the connection variable, variable name and opperation to the backend.
        get_str:
        condition_check: check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.
        connect: connect the preg with postg.
        get_weight_name: give a name for each connection weight.
        get_post_name: give a name for each post group.
        get_input_name: give a name for each input group.

    Examples:
        when building the network:
        self.connection1 = spaic.Connection(self.input, self.layer1, link_type='full')

        '''

    _connection_subclasses = dict()
    _class_label = '<con>'

    def __init__(self, pre: Assembly, post: Assembly, name=None,
                 link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn',
                 syn_kwargs=None, **kwargs):

        super(Connection, self).__init__(pre, post)
        # self.param_init = kwargs.get('param_init', None)
        self.con_kwargs = kwargs
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        if syn_kwargs is None:
            self.syn_kwargs = dict()
        else:
            self.syn_kwargs = syn_kwargs
        # assert pre._is_terminal
        # assert post._is_terminal

        self.pre = pre  # .get_assemblies()
        self.post = post  # .get_assemblies()
        self.pre_var_name = pre_var_name
        self.post_var_name = post_var_name

        self.pre_num = pre.num
        self.post_num = post.num
        # self.link_num = self.pre_num * self.post_num
        self.shape = (self.post_num, self.pre_num)

        self.pre_groups = None  # pre.get_groups()
        self.post_groups = None  # post.get_groups()
        self.pre_assemblies = None
        self.post_assemblies = None
        self.connection_information = None
        self.mask_info: List[(tuple, tuple)] = list()  # (var_code, op_code)

        self.link_type = link_type
        self.max_delay = max_delay
        self.min_delay = kwargs.get('min_delay', 0.0)
        self.sparse_with_mask = sparse_with_mask
        self._var_names = list()
        self._var_dict = dict()
        self._supers = list()
        self._link_var_codes = list()
        # self._policies = policies
        self._variables = dict()
        self._operations = list()
        self._init_operations = list()
        # self._link_op_codes = list()
        self.w_init = None
        self.b_init = None
        self.w_init_param = dict()
        self.b_init_param = dict()

        self.parameters = kwargs
        # self.bias = kwargs.get('bias', False)

        self.updated_input = kwargs.get('updated_input', True)
        self.weight_norm = kwargs.get('weight_norm', False)

        self.w_max = None
        self.w_min = None

        self.set_name(name)
        self.running_var = None
        self.running_mean = None
        self.decay = 0.9
        self._variables = dict()
        self._constant_variables = dict()

        self.flatten = kwargs.get('flatten', False)

        if len(syn_type) == 0:
            raise ValueError('Please set the param syn_type, you can set it as \'basic\'')

        if isinstance(syn_type, list):
            self.synapse_type = syn_type
        else:
            self.synapse_type = [syn_type]
        self.synapse_class = []
        self.synapse_name = []

        for i in range(len(self.synapse_type)):
            if isinstance(self.synapse_type[i], str):
                self.synapse_class.append(SynapseModel.apply_model(self.synapse_type[i]))
                self.synapse_name.append(self.synapse_type[i])  # self.model -> self.model_name
                self.synapse = []
            else:
                raise ValueError("only support set synapse model with string")

    def __new__(cls, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn',
                syn_kwargs=None, **kwargs):
        if cls is not Connection:
            return super().__new__(cls)

        if link_type in cls._connection_subclasses:
            return cls._connection_subclasses[link_type](pre, post, name, link_type, syn_type,
                                                         max_delay, sparse_with_mask, pre_var_name, post_var_name,
                                                         syn_kwargs, **kwargs)

        else:
            raise ValueError("No connection type: %s in Connection classes" % link_type)

    def get_var_names(self):
        return self._var_names

    @staticmethod
    def register(name, connection_class):
        '''
        Register a connection class. Registered connection classes can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. `'full'`)
        connection_class :
            The subclass of Connection object, e.g. an `FullConnection`, 'ConvConnection'.
        '''

        # only deal with lower case names
        name = name.lower()
        if name in Connection._connection_subclasses:
            raise ValueError(('A connection class with the name "%s" has already been registered') % name)

        if not issubclass(connection_class, Connection):
            raise ValueError(
                ('Given model of type %s does not seem to be a valid ConnectionModel.' % str(type(connection_class))))

        Connection._connection_subclasses[name] = connection_class

    def assembly_linked(self, assembly):
        if (assembly is self.pre) or (assembly is self.post):
            return True
        else:
            return False

    def replace_assembly(self, old_assembly, new_assembly):
        if old_assembly is self.pre:
            self.pre = new_assembly
        elif old_assembly is self.post:
            self.post = new_assembly
        else:
            raise ValueError("the old_assembly is not in the connnection")

    def add_super(self, assembly):
        assert isinstance(assembly, Assembly) or isinstance(assembly, Projection), "the super is not Assembly"
        self._supers.append(assembly)

    def del_super(self, assembly):
        assert assembly in self._supers, "the assembly is not in supers"
        self._supers.remove(assembly)

    def set_id(self):
        if len(self._supers) == 0:
            self.id = self.name + self.__class__._class_label
        else:
            super_ids = []
            for super in self._supers:
                if super.id is not None:
                    super_ids.append(super.id)
                else:
                    super_ids.append(super.set_id())

            self.id = self.name + self.__class__._class_label
            if len(super_ids) == 1:
                self.id = super_ids[0] + '_' + self.id
            else:
                pre_id = '/'
                for prefix in super_ids:
                    pre_id += prefix + ','
                pre_id += '/'
                self.id = pre_id + '_' + self.id
        return self.id

    def get_str(self, level):

        level_space = "" + '-' * level
        repr_str = level_space + "|name:{}, type:{}, ".format(self.name, type(self).__name__)
        repr_str += "pre:{}, ".format(self.pre.name)
        repr_str += "post:{}\n ".format(self.post.name)
        # for c in self._connections.values():
        #     repr_str += c.get_str(level)
        return repr_str

    def set_delay(self, pre_group, post_group):
        # TODO: add to unit_connections information after it changed to List[dict]

        if self.max_delay > 0:
            # print("set delay")
            # pre_num = pre_group.num
            # post_num = post_group.num
            # shape = (post_num, pre_num)
            delay_input_name, delay_output_name = self.get_delay_input_output(pre_group, post_group)

            # add delay container
            delay_queue = self._backend.add_delay(delay_input_name, self.max_delay)
            delay_name = self.get_link_name(pre_group, post_group, 'delay')
            # ONLY FOR TEST  ===============================
            ini_delay = (self.max_delay-self.min_delay) * np.random.rand(*self.shape) + self.min_delay
            # ==============================================
            self.variable_to_backend(delay_name, self.shape, ini_delay, True)

            # add delay index
            self._backend.register_standalone(Op(delay_output_name, delay_queue.select, [delay_name] ,owner=self))

            # add inital to transform initial delay_output
            self.init_op_to_backend(delay_output_name, delay_queue.transform_delay_output,
                                           [delay_input_name, delay_name])

        else:
            return

    def clamp_weight(self, weight):

        if (self.w_max is not None) and (self.w_min is not None):
            self._backend.clamp_(weight, self.w_min, self.w_max)
        elif self.w_max is not None:
            self._backend.clamp_max_(weight, self.w_max)
        elif self.w_min is not None:
            self._backend.clamp_min_(weight, self.w_min)

    def _add_label(self, var_name: str):
        if '[pre]' in var_name:
            var_name = var_name.replace("[pre]", "")
            var_name = self.get_group_name(self.pre, var_name)
        elif '[link]' in var_name:
            var_name = var_name.replace("[link]", "")
            var_name = self.get_link_name(self.pre, self.post, var_name)
        elif '[post]' in var_name:
            var_name = var_name.replace("[post]", "")
            var_name = self.get_group_name(self.post, var_name)
        elif '[input]' in var_name:
            tmp_var_name = var_name.replace("[input]", "")
            var_name = self.get_input_name(self.pre, self.post)
            if tmp_var_name not in var_name:
                raise ValueError(" the var_name tagged as [input] is not pre_var_name of the connection")
        elif var_name[0] == '[' and var_name[-1] == ']':
            var_name = var_name
        else:
            var_name = self.get_name(var_name)
        return var_name

    def add_conn_label(self, var_name: str):
        if isinstance(var_name, str):
            if '[updated]' in var_name:
                var_name = var_name.replace("[updated]", "")
                name = self._add_label(var_name)
                return name + '[updated]'
            else:
                name = self._add_label(var_name)
                return name

        elif isinstance(var_name, list) or isinstance(var_name, tuple):
            var_names = []
            for name in var_name:
                if isinstance(name, str):
                    if '[updated]' in name:
                        name = name.replace("[updated]", "")
                        name = self._add_label(name)
                        name = name + '[updated]'
                    else:
                        name = self._add_label(name)
                    var_names.append(name)
            return var_names

    @property
    def dt(self):
        return self._backend.dt

    def decode_syn_op(self, syn_ops, synapse_name, op_len):
        if op_len > 1:
            for i in range(1, op_len):
                pre_op_return_name = syn_ops[i-1][0]
                post_op_first_input = syn_ops[i][2]
                if '[updated]' in post_op_first_input:
                    post_op_first_input = post_op_first_input.replace('[updated]', '')
                if post_op_first_input == pre_op_return_name:
                    temp_name = synapse_name[i-1]+'_' + str(i-1)
                    syn_ops[i-1][0] = temp_name
                    syn_ops[i][2] = temp_name
        return syn_ops

    def decode_initializer(self, initial):
        init_name = initial.__class__.__name__
        init_param = initial.__dict__
        return init_name, init_param

    def build(self, backend):
        '''
        add the connection variable, variable name and opperation to the backend.
        '''

        self._backend = backend
        # Add weight

        self.set_delay(self.pre, self.post)

        for (key, value) in self._variables.items():
            var_name = self.add_conn_label(key)
            if isinstance(value, np.ndarray):
                if value.size > 1:
                    shape = value.shape
                else:
                    shape = ()
            elif hasattr(value, '__iter__'):
                value = np.array(value)
                if value.size > 1:
                    shape = value.shape
                else:
                    shape = ()
            else:
                shape = ()

            if 'weight' in key:
                self.variable_to_backend(name=var_name, shape=self.shape, value=value, is_parameter=self.is_parameter,
                                         is_sparse=self.is_sparse, init=self.w_init,
                                         init_param=self.w_init_param)  # (var_name, shape, value, is_parameter, is_sparse, init)
            elif 'bias' in key:
                self.variable_to_backend(name=var_name, shape=value.shape, value=value, is_parameter=self.is_parameter,
                                         init=self.b_init, init_param=self.b_init_param)
            else:
                self.variable_to_backend(var_name, shape, value=value)
            # del[self._variables[key]]
            # self._var_names.append(var_name)
            # self._var_dict[var_name] = VariableAgent(backend, var_name)

        for (key, value) in self._constant_variables.items():
            var_name = self.add_conn_label(key)
            self.variable_to_backend(name=var_name, shape=None, value=value, is_constant=True)

        weight_name = self.get_link_name(self.pre, self.post, 'weight')
        if self.sparse_with_mask:
            mask = (self.weight != 0)
            mask_name = self.get_link_name(self.pre, self.post, 'mask')
            self.variable_to_backend(mask_name, self.shape, mask)
            self.init_op_to_backend(weight_name, self.mask_operation, [weight_name, mask_name])


        if (self.w_min is not None) or (self.w_max is not None):
            self.init_op_to_backend(None, self.clamp_weight, [weight_name])

        for op in self._operations:
            addcode_op = Op(owner=self)
            addcode_op.func = op[1]
            addcode_op.output = self.add_conn_label(op[0])
            if len(op) > 3:  # 为了解决历史的单一list格式的问题
                addcode_op.input = self.add_conn_label(op[2:])
            else:
                addcode_op.input = self.add_conn_label(op[2])

            # if len(op) > 3:
            #     addcode_op[2] = addcode_op[2:]
            #     addcode_op = addcode_op[:3]
            backend.add_operation(addcode_op)

        for op in self._init_operations:
            addcode_op = []
            for ind, name in enumerate(op):
                if ind != 1:
                    addcode_op.append(self.add_conn_label(op[ind]))
                else:
                    addcode_op.append(op[ind])
            if len(op) > 3:
                addcode_op[2] = addcode_op[2:]
                addcode_op = addcode_op[:3]

            self.init_op_to_backend(addcode_op[0], addcode_op[1], addcode_op[2])

        syn_ops = []
        for i in range(len(self.synapse_class)):
            if self.synapse_class[i] is not None:
                self.synapse.append(self.synapse_class[i](self, **self.syn_kwargs))

            for (key, value) in self.synapse[i]._syn_variables.items():
                key = self.add_conn_label(key)
                self.variable_to_backend(key, value.shape, value=value)

            for (key, value) in self.synapse[i]._syn_constant_variables.items():
                key = self.add_conn_label(key)
                self.variable_to_backend(key, shape=None, value=value, is_constant=True)
            dt = backend.dt
            for (key, value) in self.synapse[i]._syn_tau_variables.items():
                key = self.add_conn_label(key)
                tau_value = np.exp(-dt / value)
                # 暂时只考虑scalar值
                self.variable_to_backend(key, shape=[1, ], value=tau_value)
            for op in self.synapse[i]._syn_operations:
                syn_ops.append(op)

        #toDO: make sure synapse with multiple ops
        # if 'basic' in self.synapse_name:
        syn_ops = self.decode_syn_op(syn_ops, self.synapse_name, len(syn_ops))

        for sop in syn_ops:
            addcode_op = Op(owner=self)
            addcode_op.func = sop[1]
            addcode_op.output = self.add_conn_label(sop[0])
            if len(sop) > 3:  # 为了解决历史的单一list格式的问题
                addcode_op.input = self.add_conn_label(sop[2:])
            else:
                addcode_op.input = self.add_conn_label(sop[2])
            backend.add_operation(addcode_op)

    @abstractmethod
    def condition_check(self, pre_group, post_group):
        NotImplemented()

    def mask_operation(self, weight, mask):
        return weight * mask

    def get_name(self, suffix_name: str):
        '''

        Args:
            pre_group(Assembly): the neuron group which needs to be connected
            post_group(Assembly): the neuron group which needs to connect with the pre_group
            suffix_name(str): represents the name of the object you want to retrieve, such as 'weight'
        Returns:
            name(str)
        '''

        name = self.id + ':' + '{' + suffix_name + '}'
        return name

    def get_link_name(self, pre_group: Assembly, post_group: Assembly, suffix_name: str):
        '''

        Args:
            pre_group(Assembly): the neuron group which needs to be connected
            post_group(Assembly): the neuron group which needs to connect with the pre_group
            suffix_name(str): represents the name of the object you want to retrieve, such as 'weight'
        Returns:
            name(str)
        '''

        name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{' + suffix_name + '}'
        return name

    def get_group_name(self, group: Assembly, suffix_name: str):
        '''
        Args:
            group(Assembly): the neuron group which needs to be connected
            suffix_name(str): represents the name of the object you want to retrieve, such as 'O'

        Returns:
            name(str)
        '''
        name = group.id + ':' + '{' + suffix_name + '}'
        return name

    # def get_post_name(self, post_group: Assembly, suffix_name: str):
    #     '''
    #     Args:
    #         post_group(Assembly): the neuron group which needs to connect with the pre_group
    #         suffix_name(str): represents the name of the object you want to retrieve, such as 'maxpool_kernel_size'
    #
    #     Returns:
    #         name(str)
    #     '''
    #     name = post_group.id + ':' + '{' + suffix_name + '}'
    #     return name

    # def get_mask_name(self, pre_group: Assembly, post_group: Assembly):
    #     name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{mask}'
    #     return name
    #
    # def get_delay_name(self, pre_group: Assembly, post_group: Assembly):
    #     name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{delay}'
    #     return name

    # 这两个特例应该要留着
    def get_delay_input_output(self, pre_group: Assembly, post_group: Assembly):
        input_name = pre_group.id + ':' + '{' + self.pre_var_name + '}'
        output_name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{' + self.pre_var_name + '}'
        return input_name, output_name

    def get_input_name(self, pre_group: Assembly, post_group: Assembly):
        '''
        Give a name for input group's output spikes,  the name consists of two parts: pre_group.id + ':' + '{0}
        Args:
            pre_group(Assembly): The neuron group which need to connect with post_group neuron.
            post_group(Assembly): the neuron group which needs to connect with the pre_group
        Returns:
            name(str)
        '''
        if self.max_delay > 0:
            name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{' + self.pre_var_name + '}'
        else:
            name = pre_group.id + ':' + '{' + self.pre_var_name + '}'
        return name

    def get_target_output_name(self, output_group: Assembly):
        name = output_group.id + ':' + '{' + self.pre_var_name + '}'
        return name

    # def Chemistry_Isyn(self):
    #     """
    #     Chemistry current synapse
    #     I = tauP*I + WgtSum
    #     """
    #     I = self.get_post_name(self.post, 'I')
    #     WgtSum = self.get_post_name(self.post, 'WgtSum')
    #     tauP = self.get_post_name(self.post, 'tauP')
    #     # I = self.get_I_ele_name(self.post)
    #     # WgtSum = self.get_target_name(self.post)
    #     # tauP = self.post.id + ':' + '{tauP}'
    #     self._syn_variables[I] = 0
    #     self._syn_variables[WgtSum] = 0
    #     self._syn_tau_constant_variables[tauP] = self.tau_p
    #
    #     self._syn_operations.append([I, 'var_linear', tauP, I, WgtSum])

    # def Electrical_synapse(self):
    #     """
    #     Electrical synapse
    #     WgtSum = weight *（V(l-1) - V(l)）
    #     """
    #     V_post = self.get_post_name(self.post, 'V')
    #     V_pre = self.get_pre_name(self.post, 'V')
    #     Vtemp1_post = self.get_link_name(self.pre, self.post, 'Vtemp')
    #     I_post = self.get_post_name(self.post, 'I')
    #     weight = self.get_link_name(self.pre, self.post, 'weight')
    #     Vtemp1_pre = self.get_link_name(self.post, self.pre, 'Vtemp')
    #     I_pre = self.get_pre_name(self.post, 'I_ele')
    #     # V_post = self.get_V_name(self.post)
    #     # V_pre = self.get_V_name(self.pre)
    #     # Vtemp1_post = self.get_Vtemp_name(self.pre, self.post)
    #     # I_post = self.get_I_ele_name(self.post)
    #     # weight = self.get_weight_name(self.pre, self.post)
    #     # Vtemp1_pre = self.get_Vtemp_name(self.post, self.pre)
    #     # I_pre = self.get_I_ele_name(self.pre)
    #
    #     self._syn_variables[Vtemp1_post] = 0
    #     self._syn_variables[I_post] = 0
    #     self._syn_variables[Vtemp1_pre] = 0
    #     self._syn_variables[I_pre] = 0
    #     self._syn_operations.append([Vtemp1_post, 'minus', V_pre, V_post])
    #     self._syn_operations.append([I_post, 'var_mult', weight, Vtemp1_post + '[updated]'])
    #     self._syn_operations.append([Vtemp1_pre, 'minus', V_post, V_pre])
    #     self._syn_operations.append([I_pre, 'var_mult', weight, Vtemp1_pre + '[updated]'])


class ConnectPolicy():

    def __init__(self, level=-1):
        super(ConnectPolicy, self).__init__()
        self.level = level
        self.policies = [self, ]

    @abstractmethod
    def checked_connection(self, new_connection: Projection):
        NotImplementedError()

    def extend_policy(self, policies):
        self.policies.extend(policies)

    def __and__(self, other):
        self.extend_policy(other.policies)
        return self

    def generate_connection(self, pre: Assembly, post: Assembly):

        candidates = []
        for p in self.policies:
            candidates.append(p.checked_connection(Projection(pre, post)))
        res = candidates[-1]
        candidates.pop()
        for con in candidates:
            res = res & con
        return res


# -*- coding: utf-8 -*-
"""
Created on 2020/9/14
@project: SPAIC
@filename: Torch_Backend
@author: Hong Chaofei
@contact: hongchf@gmail.com
@description:
"""
from .Backend import Backend, backends
import torch
import numpy as np
# from torch import fx
#from torch.nn import Module, Parameter
import torch.nn.functional as fn
import torch.nn as nn
from typing import Tuple, Dict, Callable

class Torch_Engine(torch.nn.Module):
    def __init__(self, graph_operations):
        super(Torch_Engine, self).__init__()
        self._graph_operations = graph_operations

    def forward(self, variables: Dict[str,torch.Tensor]):
        temp_dict = dict()
        update_dict = dict()
        reduce_dict = dict()

        for op in self._graph_operations:
            # for inputs
            inputs = []
            for var in op.input:
                if var[0] == 'variables_dict':
                    inputs.append(variables[var[1]])
                elif var[0] == 'temp_dict':
                    inputs.append(temp_dict[var[1]])
                elif var[0] == 'update_dict':
                    inputs.append(update_dict[var[1]])
                elif var[0] == 'reduce_dict':
                    inputs.append(reduce_dict[var[1]])
            # compute the operation
            result = op.func(*inputs)
            if len(op.output) == 1: result = [result]
            # assign the result variables
            for ind, var in enumerate(op.output):
                if var[0] == 'temp_dict':
                    temp_dict[var[1]] = result[ind]
                elif var[0] == 'update_dict':
                    update_dict[var[1]] = result[ind]
                elif var[0] == 'reduce_dict':
                    if var[1] in reduce_dict:
                        reduce_dict[var[1]].append(result[ind])
                    else:
                        reduce_dict[var[1]] = [result[ind]]

        return update_dict



class Torch_Backend(Backend):
    backend_name = 'pytorch'

    def __init__(self, device='cpu'):
        super(Torch_Backend, self).__init__()

        self.device = device[0] if isinstance(device,list) else device
        self.data_type = torch.float32
        self.debug_data = []
        self.nograd_decorator = torch.no_grad()
        self.enablegrad_decorator = torch.enable_grad()
        pass

    def build(self):
        from torch import fx
        # self._graph_var_dicts = {'variables_dict': self._variables, 'temp_dict': dict(), 'update_dict': dict(),
        #                          'reduce_dict': dict()}
        # self._graph_var_dicts['temp_dict']['example_temp_dict_pytorch_datatype'] = torch.empty(1)
        # self._graph_var_dicts['update_dict']['example_temp_dict_pytorch_datatype'] = torch.empty(1)
        # self._graph_var_dicts['reduce_dict']['example_temp_dict_pytorch_datatype'] = torch.empty(1)
        #
        #self.update_step = jit.trace(self.update_step)
        self.engine = Torch_Engine(self._graph_operations)
        self.engine = fx.symbolic_trace(self.engine)

        # self.graph_update_step = torch.jit.script(self.engine)
        self.graph_update_step = self.engine
        # print(self.engine.code)

    def build_graph(self):

        for key, value in self._InitVariables_dict.items():
            if isinstance(value, torch.Tensor):
                value = value.to(self.device)
                self._InitVariables_dict[key] = value

        super(Torch_Backend, self).build_graph()

    def to_nograd_func(self, func):
        return self.nograd_decorator(func)

    def to_grad_func(self, func):
        return self.enablegrad_decorator(func)


    # def graph_update_step(self):
    #
    #     for op in self._graph_operations:
    #         inputs = []
    #         for var in op.input:
    #             inputs.append(self._graph_var_dicts[var[0]][var[1]])
    #
    #         if op.output[0] is 'reduce_dict':
    #             self._graph_var_dicts['reduce_dict'][op.output[1]].append(op.func(*inputs))
    #         else:
    #             self._graph_var_dicts[op.output[0]][op.output[1]] = op.func(*inputs)
    #
    #     return tuple(self._graph_var_dicts['variables_dict'].values())

    #  As of now, autograd support floating point Tensor types ( half, float, double and bfloat16) and complex Tensor types (cfloat, cdouble).
    def add_backend_variable(self, module, name, shape, value=None, grad=False, is_sparse=False, init=None, init_param=None):
        '''
        Parameters
        ----------
        name
        shape
        value
        init
        Returns
        -------
        '''

        if init_param is None:
            init_param = dict()
        if value is not None:
            if hasattr(value, "__len__"):
                if tuple(value.shape) == (1,) and isinstance(value, torch.Tensor):
                    self._variables[name] = value.to(self.device) * torch.ones(shape, dtype=self.data_type, device=self.device,
                                                               requires_grad=grad)
                elif tuple(value.shape) != tuple(shape):
                    raise ValueError("Value is not scalar and the shape of Value is not equal to shape")
                # add a sparse matrices with all dimensions greater than 2
                elif is_sparse:
                    i = np.nonzero(value)
                    v = value[i]

                    # Index for sparse matrix
                    sparse_index = name + '_sparse_index'
                    self._variables[sparse_index] = torch.LongTensor(i).to(device=self.device)
                    self._InitVariables_dict[sparse_index] = self._variables[sparse_index]

                    # Value for sparse matrix
                    sparse_value = name + '_sparse_value'
                    if init is not None:
                        # self._variables[sparse_value] = self.init_param(True, init)
                        data = torch.empty(shape, dtype=self.data_type, device=self.device, requires_grad=True)
                        self._variables[sparse_value] = self.param_init_operate[init](data, **init_param)
                    else:
                        self._variables[sparse_value] = torch.tensor(v, dtype=self.data_type, requires_grad=True, device=self.device)
                    self._parameters_dict[sparse_value] = self._variables[sparse_value]

                    # The shape of sparse matrix
                    sparse_shape = name + '_sparse_shape'
                    self._variables[sparse_shape] = torch.Size(shape)
                    self._InitVariables_dict[sparse_shape] = self._variables[sparse_shape]

                    # Sparse matrix
                    self._variables[name] = torch.sparse.FloatTensor(self._variables[sparse_index], self._variables[sparse_value], self._variables[sparse_shape])
                else:
                    # add a non sparse matrices with all dimensions greater than 2
                    if init is not None:
                        data = torch.empty(shape, dtype=self.data_type, device=self.device, requires_grad=grad)
                        init = init.lower()
                        if init in self.param_init_operate.keys():
                            self._variables[name] = self.param_init_operate[init](data, **init_param)
                        else:
                            raise ValueError("No initialize method: %s in param_init_operate" % init)
                    else:
                        if isinstance(value, torch.Tensor):
                            self._variables[name] = value.clone().detach().to(self.device)
                        else:
                            self._variables[name] = torch.tensor(value, dtype=self.data_type, device=self.device,
                                                                   requires_grad=grad)

            elif len(shape) == 0:
                # add constant
                self._variables[name] = torch.tensor(value, dtype=self.data_type, device=self.device, requires_grad=grad)

            else:
                # add a matrix through constant
                if init is not None:
                    # self._variables[name] = self.init_param(grad, init)
                    data = value*torch.ones(shape, dtype=self.data_type, device=self.device, requires_grad=grad)
                    init = init.lower()
                    if init in self.param_init_operate.keys():
                        self._variables[name] = self.param_init_operate[init](data, **init_param)
                    else:
                        raise ValueError("No initialize method: %s in param_init_operate" % init)
                else:
                    self._variables[name] = value*torch.ones(shape, dtype=self.data_type, device=self.device, requires_grad=grad)

        return self._variables[name]

    def set_variable_value(self, name, value, is_parameter):
        if is_parameter:
            assert name in self._parameters_dict
            assert self._parameters_dict[name].shape == value.shape
            if not isinstance(value, torch.Tensor):
                value = torch.tensor(value, dtype=self._parameters_dict[name].dtype, device=self._parameters_dict[name].device)
            with torch.no_grad():
                self._parameters_dict[name].data = value
        else:
            assert name in self._InitVariables_dict
            assert self._backend._InitVariables_dict[name].shape == value.shape
            if not isinstance(value, torch.Tensor):
                value = torch.tensor(value, dtype=self._InitVariables_dict[name].dtype, device=self._InitVariables_dict[name].device)
            with torch.no_grad():
                self._InitVariables_dict[name].data = value


    # def init_param(self, grad, *init):
    #     if init[0] in self.param_init_operate:
    #         init_op = self.param_init_operate[init[0]]
    #     else:
    #         raise ValueError("No init operate %s in param_init_operate" % init[0])
    #     inputs = []
    #     shape = init[1]
    #     data = torch.empty(shape, dtype=self.data_type, device=self.device, requires_grad=grad)
    #     inputs.append(data)
    #
    #     for var in init[2:]:
    #         inputs.append(var)
    #     return init_op(*inputs)

    def sparse_to_dense(self, index_name, value_name, shape_name):
        return torch.sparse.FloatTensor(self._variables[index_name], self._variables[value_name], self._variables[shape_name])

    def get_str(self, level):
        return level*' ' + 'torch_backend'

    def threshold(self, x, v_th):
        return torch.gt(x, v_th).type(self.data_type)

    def reset(self, v, o):
        return o.eq(0)*v

    def cat(self, x, dim=1):
        return torch.cat(x, dim)

    def stack(self, x, dim=1):       # 在指定维度dim上连接（concatenate）若干个张量。
        try:
            return torch.stack(x, dim)
        except:
            # patch for SLIF 2[O]
            for ii in range(len(x)):
                if x[ii].dim() ==2:
                    tmp = torch.zeros_like(x[ii])
                    tmp = torch.stack([x[ii], tmp], dim=1)
                    x[ii] = tmp
            return torch.stack(x, dim)

    def reduce_sum(self, x, *dim):
        if len(dim)==0:
            dim = 1
        return torch.sum(x, dim=dim)

    def index_select(self, x,  indices, dim=1):
        return torch.index_select(x, dim=dim, index=indices)

    def permute(self, x, permute_dim):
        return x.permute(permute_dim)

    def view(self, x, view_dim):

        x = x.contiguous().view(view_dim)
        return x


    def scatter(self, x, indices):
        return torch.scatter(x, dim=0, index=indices)

    def conv1d(self, x, kernel):
        return torch.conv1d(x, kernel)

    def conv_trans1d(self, x, kernel):
        return torch.conv_transpose1d(x, kernel)

    def conv_2d(self, x, kernel, stride, padding, dilation, groups):
        if x.dim() == kernel.dim() + 1:
            xshape = list(x.shape)
            xshape[0] = xshape[0]*xshape[1]
            extend_size = xshape[1]
            xshape.pop(1)
            out = fn.conv2d(x.reshape(xshape), kernel, stride=stride, padding=padding, dilation=dilation, groups=groups)
            outshape = list(out.shape)
            outshape[0] = outshape[0]//extend_size
            outshape.insert(1, extend_size)
            return out.view(outshape)
        else:
            return fn.conv2d(x, kernel, stride=stride, padding=padding, dilation=dilation, groups=groups)

    def conv_2d_complex(self, x, kernel, stride, padding, dilation, groups, beta):
        if x.dtype.is_complex:
            x = beta ** x.imag * (x.real * (0 + 1.0j))
        else:
            x = x*(0+1.0j)
        real = fn.conv2d(x.real, kernel, stride=stride, padding=padding, dilation=dilation, groups=groups)
        imag = fn.conv2d(x.imag, kernel, stride=stride, padding=padding, dilation=dilation, groups=groups)
        return torch.complex(real, imag)



    def conv_max_pool2d(self, x, kernel, pool_kernel, stride, pool_stride, padding, pool_padding, dilation, groups):
        return fn.max_pool2d(fn.conv2d(x, kernel, stride=stride, padding=padding,
                             dilation=dilation, groups=groups), kernel_size=pool_kernel,
                             stride=pool_stride, padding=pool_padding)

        # return fn.conv2d(fn.max_pool2d(x, int(max_kernel[0])), kernel, stride=int(stride), padding=int(padding), dilation=int(dilation), groups=int(groups))

    def conv_avg_pool2d(self, x, kernel, pool_kernel, stride, pool_stride, padding, pool_padding, dilation, groups):
        return fn.avg_pool2d(fn.conv2d(x, kernel, stride=stride, padding=padding,
                             dilation=dilation, groups=groups), kernel_size=pool_kernel,
                             stride=pool_stride, padding=pool_padding)
    def conv_add_bias(self, x, bias):
        bias_t = bias.repeat(x.shape[-2], x.shape[-1], 1).permute(2, 1, 0)
        return x+bias_t
    def max_pool2d(self, x, pool_kernel, pool_stride, pool_padding):
        return fn.max_pool2d(x, kernel_size=pool_kernel, stride=pool_stride, padding=pool_padding)
    def post_max_pool2d_complex(self, x, pool_kernel, pool_stride, pool_padding):
        pool_imag, pool_index = fn.max_pool2d(x.imag, kernel_size=pool_kernel, return_indices=True,
                                              stride=pool_stride, padding=pool_padding)
        x_shape = x.shape
        pool_shape = pool_index.shape
        pool_real = torch.gather(x.real.view(x_shape[0], x_shape[1], -1), dim=-1,
                                 index=pool_index.view(x_shape[0], x_shape[1],-1)).view(pool_shape[0], pool_shape[1],
                                                                                        pool_shape[2], pool_shape[3])


        return torch.complex(real=pool_real, imag=pool_imag)

    def avg_pool2d(self, x, pool_kernel, pool_stride, pool_padding):
        return fn.avg_pool2d(x, kernel_size=pool_kernel, stride=pool_stride, padding=pool_padding)

    def batchnorm2d(self, x, num_features):
        # 该实现方式忽略了running_mean 和 running_var
        # 当 batch_size 较小时，在推理阶段的统计特性就会和全局统计特性有着较大偏差，从而导致糟糕的效果，这种情况下推荐使用
        # SPAIC 的 Module 模块实现 batchnorm。
        device = x.device
        bn_2d = torch.nn.BatchNorm2d(num_features).to(device=device)
        return bn_2d(x)

    def dropout(self, x, p, inplace=False):
        return fn.dropout(x, p=p, inplace=inplace)

    def reshape_mat_mult(self, A, X):

        if A.dim() == 4:
            (batchsize, outchannels, H, W) = A.shape
            A = A.view(batchsize, -1)
        elif A.dim() == 5:
            (batchsize, extend, outchannels, H, W) = A.shape
            A = A.view(batchsize, extend, -1)

        return torch.matmul(A, X.permute(1, 0))

    def im2col_indices(self, x, kh, kw, padding, stride):
        return fn.unfold(x, (kh, kw), padding=padding, stride=stride)

    def conv2d_flatten(self, x):
        return x.view(x.shape[0], x.shape[1], -1)

    def feature_map_flatten(self, x):
        return x.view(x.shape[0], -1)

    def add(self, x, y):

        return x + y

    def minus(self, x, y):
        return x - y

    def div(self, x, y):
        return torch.div(x, y)

    def relu(self, x):
        return torch.relu(x)

    def sigmoid(self, x):
        return torch.sigmoid(x)

    def mat_mult_weight(self, A, X):
        '''
        Parameters
        ----------
        A--->preGroup:input
        X--->postGroup:weight
        Returns
        -------
        '''

        X = X.permute(1, 0)
        return torch.matmul(A, X)

    def mat_mult_weight_complex(self, A, X, beta):
        '''
        Parameters
        ----------
        A--->preGroup:input
        X--->postGroup:weight
        beta---> postGroup:beta_complex
        Returns
        -------
        '''
        if A.dtype.is_complex:
            A = A.unsqueeze(-2)
            beta = beta.unsqueeze(-1)
            real = A.real
            imag = A.imag
            O = beta ** imag * (real * (0 + 1.0j))
            return torch.sum(O * X, dim=-1)
        else:
            X = X.permute(1, 0)
            Out = torch.matmul(A.to(X.dtype), X)
            Out = Out*(0.0+1.0j)
            return Out


    def mat_mult_pre(self, A, X):
        '''
        Parameters
        ----------
        A--->preGroup:input
        X--->postGroup:weight
        Returns
        -------
        '''
        A = A.permute(1, 0)
        return torch.matmul(A, X)

    def mat_mult(self, A, X):
        '''
        Parameters
        ----------
        A--->preGroup:input
        X--->postGroup:weight
        Returns
        -------
        '''
        return torch.matmul(A, X)

    def bmm(self, A, X):
        '''
        Parameters
        ----------
        A---> postGroup
        X---> preGroup
        Returns
        -------
        '''
        return torch.bmm(A, X)

    def ger(self, A, X):
        '''
        Parameters
        ----------
        A---> postGroup
        X---> preGroup
        Returns
        -------
        '''
        return torch.ger(A, X)

    def sparse_mat_mult_weight(self, A, X):
        '''
       Parameters
       ----------
       A--->preGroup:sparseWeight(post, pre)
       X--->postGroup:input(batch, pre)
       Returns
       -------
       '''
        X = X.permute(1, 0)
        result = torch.sparse.mm(A, X)
        result = result.permute(1, 0)
        return result

    def var_mult(self, A, X):
        return A * X

    def mult_sum_weight(self, A, X):
        # X = X.permute(1, 0)
        A = A.permute(0, 2, 1)
        return torch.sum(A * X, dim=-1)

    def mat_linear(self, A, X, b):
        return torch.matmul(A, X) + b

    def var_linear(self, A, X, b):

        return A*X+b

    def unsqueeze(self, X, dim):
        return torch.unsqueeze(X, dim)

    def to_numpy(self, data: torch.Tensor):
        return data.detach().cpu().numpy()

    def to_tensor(self, data):
        if isinstance(data, torch.Tensor):
            return data.to(torch.float).to(self.device)
        else:
            return torch.tensor(data, dtype=torch.float, device=self.device)

    def exp(self, x):
        return torch.exp(x)


    def clamp_(self, data, min, max):
        with torch.no_grad():
            data.clamp_(min, max)


    def clamp_max_(self, data, max):
        with torch.no_grad():
            data.clamp_max_(max)



    def clamp_min_(self, data, min):
        with torch.no_grad():
            data.clamp_min_(min)


    def uniform(self, data, a=-0.0, b=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            a(float): the lower bound of the uniform distribution
            b(float): the upper bound of the uniform distribution
        Returns:
            torch.nn.init.uniform_(data, a=0.0, b=1.0)
        '''
        return torch.nn.init.uniform_(data, a, b)


    def normal(self, data, mean=0.0, std=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            mean(float): the mean of the normal distribution
            std(float): the standard deviation of the normal distribution
        Returns:
            torch.nn.init.normal_(data, mean=0.0, std=1.0)
        '''
        return torch.nn.init.normal_(data, mean, std)



    def xavier_normal(self, data, gain=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            gain: an optional scaling factor
        Returns:
            torch.nn.init.xavier_normal_(data, gain=1.0)
        '''
        return torch.nn.init.xavier_normal_(data, gain)


    def xavier_uniform(self, data, gain=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            gain: an optional scaling factor
        Returns:
            torch.nn.init.xavier_uniform_(data, gain=1.0)
        '''
        return torch.nn.init.xavier_uniform_(data, gain)


    def kaiming_normal(self, data, a=0, mode='fan_in', nonlinearity='leaky_relu'):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            a: the negative slope of the rectifier used after this layer (only used with 'leaky_relu')
            mode: either 'fan_in' (default) or 'fan_out'. Choosing 'fan_in' preserves the magnitude of the variance of the weights in the forward pass. Choosing 'fan_out' preserves the magnitudes in the backwards pass.
            nonlinearity: the non-linear function (nn.functional name), recommended to use only with 'relu' or 'leaky_relu' (default).
        Returns:
            torch.nn.init.kaiming_normal_(data, a=0, mode='fan_in', nonlinearity='leaky_relu')
        '''
        return torch.nn.init.kaiming_normal_(data, a, mode, nonlinearity)


    def kaiming_uniform(self, data, a=0, mode='fan_in', nonlinearity='leaky_relu'):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            a: the negative slope of the rectifier used after this layer (only used with 'leaky_relu')
            mode: either 'fan_in' (default) or 'fan_out'. Choosing 'fan_in' preserves the magnitude of the variance of the weights in the forward pass. Choosing 'fan_out' preserves the magnitudes in the backwards pass.
            nonlinearity: the non-linear function (nn.functional name), recommended to use only with 'relu' or 'leaky_relu' (default).
        Returns:
            torch.nn.init.kaiming_uniform_(data, a=0, mode='fan_in', nonlinearity='leaky_relu')
        '''
        return torch.nn.init.kaiming_uniform_(data, a, mode, nonlinearity)

    def constant(self, data, constant_value=0.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            constant_value(float): the value to fill the tensor with
        Returns:
            torch.nn.init.constant_(data, constant_value)
        '''
        return torch.nn.init.constant_(data, constant_value)

    def sparse(self, data, sparsity=0.1, std=0.01):
        '''
        Args:
            data(tensor): an n-dimensional `torch.Tensor`
            sparsity(float): The fraction of elements in each column to be set to zero
            std(float): the standard deviation of the normal distribution used to generate
            the non-zero values
        Returns:
            torch.nn.init.sparse_(data, sparsity, std)
        '''
        return torch.nn.init.sparse_(data, sparsity, std)

    def weight_norm(self, weight, amp):
        w_norm = torch.norm(weight, p=2, dim=1, keepdim=True)
        # print(amp.item(), w_norm.item())
        return weight*amp/w_norm


    def sin(self, x):
        return torch.sin(x)

    def cos(self, x):
        return torch.cos(x)

    def tan(self, x):
        return torch.tan(x)

    def log(self, x):
        return torch.log(x)

    def log2(self, x):
        return torch.log2(x)

    def log10(self, x):
        return torch.log10(x)

    # def reset(self, x, v_reset, u_reset, spike):
    #
    #     # if hasattr(x, "__len__"):
    #     #     if x.shape != spike.shape:
    #     #         raise ValueError("%s and %s do not match" % (x.shape, spike.shape))
    #     mask = torch.eq(spike, 1)
    #     x[mask] = v_reset
    #     x[mask] += u_reset
    #     return x

    # def izh_v(self, v, u, psp):
    #     v = v+self.dt*(0.04*v*v+5*v+140-u+psp)
    #     return v
    #
    # def izh_u(self, a, b, v, u):
    #     u = u+self.dt*a*(b*v-u)
    #     return u


backends[Torch_Backend.backend_name] = Torch_Backend





# test = Torch_Backend()
# th = test.basic_operate['threshold']
# print(th(-1.0))


# -*- coding: utf-8 -*-
"""
Created on 2020/8/17
@project: SPAIC
@filename: Network_loader
@author: Mengxiao Zhang
@contact: mxzhangice@gmail.com

@description:
对已按格式储存网络的加载和重建
"""

import yaml
import json

import spaic
import torch


def network_load(filename=None, path=None, device='cpu', load_weight=True):
    '''
        The main function for getting the target filename and reloading the
            network.

        Args:
            filename(str) : The filename of the target network, given by user.
            dataloader(dataloader) : The dataloader for input node layer,
                should be given or crash.
            encoding(str) : The encoding model chosen by user.
            device(str) : The device type we choose to run our network.

        Return:
            net(Assembly) :The network that reloaded from the file.

        Example:
            Net = network_load('TestNetwork', dataloader, 'poisson')

    '''
    import os
    if path:
        filedir = path + '/' + filename
    else:
        path = './'
        # filedir = path + filename
    file = filename.split('.')[0]
    origin_path = os.getcwd()
    os.chdir(path+'/'+file)
    if os.path.exists(filename):
        with open(filename, 'r') as f:
            data = f.read()
            if data.startswith('{'):
                data = json.loads(data)
            else:
                data = yaml.load(data, Loader=yaml.FullLoader)

    else:
        if os.path.exists(filename+'.yml'):
            with open(filename+'.yml', 'r') as f:
                data = yaml.load(f, Loader=yaml.FullLoader)

        elif os.path.exists(filename+'.json'):
            with open(filename+'.json', 'r') as f:
                data = json.load(f)

        elif os.path.exists(filename+'.txt'):
            with open(filename+'.txt', 'r') as f:
                data = f.read()
                if data.startswith('{'):
                    data = json.loads(data)
                else:
                    data = yaml.load(data, Loader=yaml.FullLoader)

        else:
            raise ValueError("file %s doesn't exist, please check the "
                             "filename" % filename)

    net = ReloadedNetwork(net_data=data, device=device, load_weight=load_weight)

    os.chdir(origin_path)
    return net


class ReloadedNetwork(spaic.Network):
    '''
        The network rebuild from the yaml file.

        Args:
            net_data(dict) : The network information reloaded from yaml files.
            dataloader(dataloader) : The dataloader of input layer, since the
                large scale of data, we will not save the data.
            encoding(str) : The encoding model, default as poisson, will change
                in the future.
            backend(backend) : Backend that user want to use.
            learner(str) : The learning model of this network, will change in
                the future.
            learner_alpha(int) : The parameter alpha for learning model, will
                change in the future.
            device(str) : The type of device that run our model.

        Methods:
            load_net(self, data: dict) : The function for load the whole
                network, main function of this class.
            load_layer(self, layer: dict) : The function for load layer.
            load_connection(self, con: dict) : The function for load
                connection.
            load_node(self, node: dict) : The function for load node like input
                or output.
            load_backend(self, path: str): The function for load backend.

        Example:
            Net = ReloadNetwork(net_data, dataloader, 'poisson', backend,
                'STCA', 0.5)

    '''
    def __init__(self, net_data: dict, backend=None, device='cpu', load_weight=True, data_type=None):
        super(ReloadedNetwork, self).__init__()

        self.device = device
        self.name = list(net_data)[0]
        self._backend_info = []

        self.load_net(net_data)
        if backend is None:
            backend = spaic.Torch_Backend(device)

        self.set_backend(backend)
        self.set_backend_data_type()
        # self._learner = Learner(algorithm='STCA', lr=0.5, trainable=self)
        self.build()
        if load_weight:
            self.load_backend(device)

        del self._backend_info

    def load_net(self, data):
        '''
            The function for load the whole network, main function of this class.

            Args:
                data(dict) : The data should contains the network structure and
                    parameter from yaml.

        '''
        setid = 0
        data = data[list(data)[0]]
        for g in data:
            if list(g)[0] == 'monitor':
                monitors = g.get('monitor')
                for monitor in monitors:
                    self.load_monitor(monitor)
                continue
            if list(g)[0] == 'backend':
                self._backend_info = g[list(g)[0]]
                continue
            para = g[list(g)[0]]
            if type(para) is dict:
                if para.get('_class_label') == '<neg>':
                    lay_name = para.get('name')
                    self.add_assembly(name=lay_name,
                                      assembly=self.load_layer(para))
                elif para.get('_class_label') == '<nod>':
                    nod_name = para.get('name')
                    self.add_assembly(name=nod_name,
                                      assembly=self.load_node(para))
                elif para.get('_class_label') == '<con>':
                    con_name = para.get('name')
                    self.add_connection(name=con_name,
                                        connection=self.load_connection(pnet=self, con=para))
                elif para.get('_class_label') == '<prj>':
                    prj_name = para.get('name')
                    self.add_projection(name=prj_name,
                                        projection=self.load_projection(prj=para))
                elif para.get('_class_label') == '<learner>':
                    learner = self.load_learner(para)
                    self._learners[para.get('name')] = learner
                else:
                    print('Unknown class label %d' % para['_class_label'])

                    break
            else:
                self.add_assembly(name=list(g)[0], assembly=self.load_assembly(list(g)[0], para))


    def load_assembly(self, name, assembly: list):
        target = spaic.Assembly(name=name)
        for g in assembly:
            para = g[list(g)[0]]
            if para.get('_class_label') == '<neg>':
                lay_name = para.get('name')
                target.add_assembly(name=lay_name,
                                  assembly=self.load_layer(para))
            elif para.get('_class_label') == '<nod>':
                nod_name = para.get('name')
                target.add_assembly(name=nod_name,
                                  assembly=self.load_node(para))
            elif para.get('_class_label') == '<con>':
                con_name = para.get('name')
                target.add_connection(name=con_name,
                                    connection=self.load_connection(pnet=target, con=para))
            elif para.get('_class_label') == '<prj>':
                prj_name = para.get('name')
                target.add_projection(name=prj_name,
                                    projection=self.load_projection(prj=para))
        return target

    def load_layer(self, layer: dict):
        '''
            The function for load layer.

            Args:
                layer(dict): Data contains the parameters of layers.

            Return：
                NeuronGroup with need parameters.

        '''
        # layer.pop('_class_label')
        # parameters = self.trans_para(layer.get('parameters'))
        parameters = layer.get('parameters')
        return_neuron = spaic.NeuronGroup(
            num= layer.get('num', 100),
            shape= layer.get('shape', [100]),
            neuron_type     = layer.get('type', 'non_type'),
            neuron_position = layer.get('position', 'x, y, z'),
            model= layer.get('model_name', 'clif'),
            name            = layer.get('name'),
            **parameters
        )
        return_neuron.id = layer.get('id', None)
        return return_neuron

    @staticmethod
    def load_connection(pnet, con: dict):
        '''
            The function for load connections,

            Args:
                con(dict): Data contains the parameters of connections.

            Return:
                Connection with needed parameters.

        '''
        # con.pop('_class_label')

        # if con['pre'] in net._groups.keys() and \
        #         con['post'] in net._groups.keys():
        #     con['pre']  = net._groups[con['pre']]
        #     con['post'] = net._groups[con['post']]
        # else:
        #     print("Trans_error")
        #     print(net._groups.keys())
        if pnet._class_label == '<prj>':
            for pretarget in pnet.pre.get_groups():
                if con['pre'] == pretarget.id:
                    con['pre'] = pretarget
            for posttarget in pnet.post.get_groups():
                if con['post'] == posttarget.id:
                    con['post'] = posttarget
        else:
            for target in pnet.get_groups():
                if con['pre'] == target.id:
                    con['pre'] = target
                if con['post'] == target.id:
                    con['post'] = target

        assert not isinstance(con['pre'], str)
        assert not isinstance(con['post'], str)

        # con.pop('weight_path')
        return_conn = spaic.Connection(
            pre              = con.get('pre'),
            post             = con.get('post'),
            name             = con.get('name'),
            link_type        = con.get('link_type', 'full'),
            syn_type         = con.get('synapse_type', ['basic_synapse']),
            max_delay        = con.get('max_delay', 0),
            sparse_with_mask = con.get('sparse_with_mask', False),
            pre_var_name     = con.get('pre_var_name', 'O'),
            post_var_name    = con.get('post_var_name', 'WgtSum'),
            **con.get('parameters')
        )
        return_conn.id = con.get('id', None)
        return return_conn

    def load_projection(self, prj: dict):
        '''
            The function for load projection,

            Args:
                prj(dict): Data contains the parameters of projection.

            Return:
                Projection with needed parameters.

        '''
        if prj['pre'] in self._groups.keys() and \
                prj['post'] in self._groups.keys():
            prj['pre']  = self._groups[prj['pre']]
            prj['post'] = self._groups[prj['post']]
        else:
            print("Trans_error")
            print(self._groups.keys())

        assert not isinstance(prj['pre'], str)
        assert not isinstance(prj['post'], str)

        this_prj = spaic.Projection(
            pre                  = prj.get('pre'),
            post                 = prj.get('post'),
            name                 = prj.get('name'),
            link_type            = prj.get('link_type', 'full'),
            # policies             = prj.get('policies', []),
            ConnectionParameters = prj.get('ConnectionParameters'),
        )

        for conn in prj['conns']:
            for key, value in conn.items():
                this_prj.add_connection(
                    con=self.load_connection(pnet=this_prj, con=value),
                    name=key,
                )


        # prj['policies'] = []
        # from spaic.Network.ConnectPolicy import IndexConnectPolicy, ExcludedTypePolicy, IncludedTypePolicy
        # policy_dict = {'Included_policy': IncludedTypePolicy,
        #                'Excluded_policy': ExcludedTypePolicy}
        #
        # for ply in prj['_policies']:
        #     if ply['name'] == 'Index_policy':
        #         prj['policies'].append(IndexConnectPolicy(pre_indexs=ply['pre_indexs'],
        #                                                   post_indexs=ply['post_indexs'],
        #                                                   level=ply['level']))
        #     else:
        #         prj['policies'].append(policy_dict[ply['name']](pre_types=ply['pre_types'],
        #                                                   post_types=ply['post_types'],
        #                                                   level=ply['level']))

        # con.pop('weight_path')
        return this_prj

    def load_node(self, node: dict):
        '''
            The function for load node like input or output.

            Args:
                node(dict): Data contains the parameters of nodes.

            Return:
                Node of input or output layer, contains needed parameters.

        '''

        Node_dict = {'decoder': spaic.Decoder, 'action': spaic.Action, 'reward': spaic.Reward,
                     'generator': spaic.Generator, 'encoder': spaic.Encoder}

        if node.get('kind') == 'decoder':
            return_node = Node_dict[node.get('kind')](
                num             = node.get('num'),
                dec_target      = self._groups.get(node.get('dec_target', None), None),
                dt              = node.get('dt', 0.1),
                # time            = node.get('time'),
                coding_method   = node.get('coding_method', 'poisson'),
                coding_var_name = node.get('coding_var_name', 'O'),
                node_type       = node.get('type', None),
                **node.get('coding_param')
            )
        else:
            return_node = Node_dict[node.get('kind')](
                shape           = node.get('shape', None),
                num             = node.get('num'),
                dec_target      = self._groups.get(node.get('dec_target', None), None),
                dt              = node.get('dt', 0.1),
                # time            = node.get('time'),
                coding_method   = node.get('coding_method', 'poisson'),
                coding_var_name = node.get('coding_var_name', 'O'),
                node_type       = node.get('type', None),
                **node.get('coding_param')
            )
        return_node.id = node.get('id', None)
        return return_node

    def load_backend(self, device):
        '''
            The function for load backend parameters.

        '''

        # key_parameters_dict = ['_variables', '_parameters_dict', '_InitVariables_dict']
        key_parameters_dict = ['_parameters_dict']
        key_parameters_list = ['dt', 'runtime', 'time', 'n_time_step']
        typical = ['_graph_var_dicts']


        import torch
        # import os

        if self._backend_info:
            # self._backend.data_type
            for key in key_parameters_list:
                self._backend.__dict__[key] = self._backend_info[key]

            # for key in key_parameters_dict:
            path = self._backend_info['_parameters_dict']
            data = torch.load(path)
            for key, value in data.items():
                # print(key, 'value:', value)
                self._backend._parameters_dict[key] = value.to(device)
            # self._backend.data_type = value.dtype
        # #
        # for key, value in self._backend.__dict__['_parameters_dict'].items():
        #     self._backend.__dict__['_variables'][key] = value  # 这些变量的 requires_grad应该都是True

        return

    def set_backend_data_type(self, data_type=None):
        import torch
        supported_data_type = {'torch.float64': torch.float64,
                               'torch.float32': torch.float32,
                               'torch.float16': torch.float16,
                               'torch.bfloat16': torch.bfloat16,
                               'torch.int64': torch.int64,
                               'torch.int32': torch.int32,
                               'torch.int16': torch.int16,
                               'torch.bool': torch.bool,
                               'torch.uint8': torch.uint8}
        if data_type:
            self._backend.data_type = data_type
        else:
            if self._backend_info:
                self._backend.data_type = supported_data_type[self._backend_info['data_type']]

    def load_learner(self, learner: dict):
        '''
            The function for load learners' parameters.

        '''
        if '<net>' in learner['trainable']:  ## If self in net, use the whole net as the trainable traget.
            learner.pop('trainable')
            builded_learner = spaic.Learner(
                algorithm = learner.get('algorithm'),
                trainable = self,
                **learner.get('parameters')
            )
        else:
            trainable_list = []
            for trains in learner['trainable']:
                if trains in self._groups:
                    trainable_list.append(self._groups[trains])
                elif trains in self._connections:
                    trainable_list.append(self._connections[trains])
            learner.pop('trainable')
            if learner.get('parameters'):
                builded_learner = spaic.Learner(
                    trainable = trainable_list,
                    algorithm = learner.get('algorithm'),
                    **learner.get('parameters')
                    )
            else:
                builded_learner = spaic.Learner(
                    trainable = trainable_list,
                    algorithm = learner.get('algorithm')
                )
        if learner.get('optim_name', None):
            builded_learner.set_optimizer(optim_name=learner.get('optim_name'),
                                      optim_lr=learner.get('optim_lr'),
                                      **learner.get('optim_para'))
        if learner.get('lr_schedule_name', None):
            builded_learner.set_schedule(lr_schedule_name=learner.get('lr_schedule_name'),
                                         **learner.get('lr_schedule_para'))

        return builded_learner

    def load_monitor(self, monitor: dict):
        '''
        Used to add monitors to the model according to the

        Args:
            monitor: a dict that contains monitors' information.


        '''
        monitor_dict = {'StateMonitor': spaic.StateMonitor,
                        'SpikeMonitor': spaic.SpikeMonitor}

        for name, mon in monitor.items():
            for target in self.get_groups():
                if mon['target'] == target.id:
                    mon['target'] = target
                    break
            for target in self.get_connections():
                if mon['target'] == target.id:
                    mon['target'] = target
                    break

            self.add_monitor(name=name,
                                 monitor=monitor_dict[mon.get('monitor_type', 'StateMonitor')](
                                 target=mon['target'],
                                 var_name=mon['var_name'],
                                 dt=mon['dt'],
                                 get_grad=mon['get_grad'],
                                 nbatch=mon['nbatch'],
                                 index=mon['index']))

    def trans_para(self, para):
        if isinstance(para, dict):
            for key, value in para.items():
                para[key] = self.trans_para(value)
        else:
            para = torch.tensor(para, dtype=torch.float32, device=self.device)
        return para


# -*- coding:  utf-8 -*-
"""
Created on 2020/8/5
@project: SPAIC
@filename: Connection
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
定义神经集群间的连接，包括记录神经元集群、连接的突触前、突触后神经元编号、连接形式（全连接、稀疏连接、卷积）、权值、延迟 以及连接产生函数、重连接函数等。
"""
from ..Network.Topology import Connection
from ..Network.Assembly import Assembly
import numpy as np
import scipy.sparse as sp
import torch
from ..IO.Initializer import BaseInitializer
import random

class FullConnection(Connection):

    '''
    each neuron in the first layer is connected to each neuron in the second layer.

    Args:
        pre(Assembly): The assembly which needs to be connected
        post(Assembly): The assembly which needs to connect the pre
        link_type(str): full
    '''

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(FullConnection, self).__init__(pre=pre, post=post, name=name,
                                             link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                             sparse_with_mask=sparse_with_mask,
                                             pre_var_name=pre_var_name, post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.005)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

        if weight is None:
            # Connection weight
            self.weight = self.w_std*np.random.randn(*self.shape) + self.w_mean
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                self.weight = np.empty(self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight


        self.Isyn_bn = kwargs.get('Isyn_bn', False)
        if self.Isyn_bn is True:
            self.Is = []
            self.beta_ave = kwargs.get('beta_ave', 0.99)
            self.beta_std = kwargs.get('beta_std', 0.999)
            self.beta_bn = kwargs.get('beta_bn', 0.001)
            self.targ_ave = kwargs.get('targ_ave', 1.5)
            self.targ_std = kwargs.get('targ_std', 2.0)
            self.running_ave = self.targ_ave
            self.running_std = self.targ_std
            self.running_bias = 0.0
            self.running_scale = 1.0

            self._operations.append([None, self.update_ave, 'V'+'[post]'])
            self._init_operations.append(['weight[link]', self.update_bn, [post_var_name+'[post]', 'weight[link]']])

        # construct unit connection information by policies,
        # construct in __init__ is potentially bad, as network structure may change before build. should add new function
        # self.connection_inforamtion = ConnectInformation(self.pre, self.post)
        # self.connection_inforamtion.expand_connection()
        # for p in self._policies:
        #     self.connection_inforamtion = self.connection_inforamtion & p.generate_connection(self.pre, self.post)

    def update_ave(self, Isyn):
        with torch.no_grad():
            if Isyn.dtype.is_complex:
                self.Is.append(Isyn.imag.detach())
            else:
                self.Is.append(Isyn.detach())

    def update_bn(self, Isyn, weight):
        if self.training:
            weight.register_hook(self.norm_hook)
            with torch.no_grad():
                if len(self.Is) > 1:
                    Is = torch.cat(self.Is, dim=0)
                    mean_rate = torch.mean(Is, dim=0)
                    std = torch.std(Is, dim=0)
                    self.running_ave = self.beta_ave*self.running_ave + (1-self.beta_ave)*mean_rate
                    self.running_std = self.beta_std*self.running_std + (1-self.beta_std)*std

                    d_bias = (self.targ_ave - self.running_ave)*torch.sigmoid(20.0*(self.targ_ave - self.running_ave)*(self.running_ave-mean_rate))
                    self.running_bias = self.running_bias + self.beta_bn*d_bias
                    d_scale = torch.exp((self.targ_std - self.running_std)) - 1.0
                    self.running_scale = self.running_scale + self.beta_bn*d_scale
                    self.Is = []
                    self.weight_change = True
                else:
                    self.weight_change = False
                    self.running_scale = torch.ones_like(torch.mean(Isyn.real, dim=0))
                    self.running_bias = torch.zeros_like(torch.mean(Isyn.real, dim=0))
        else:
            self.Is = []

        if self.weight_change:
            weight = self.running_scale.unsqueeze(-1) * weight + self.running_bias.unsqueeze(-1)
            if weight.requires_grad:
                weight.retain_grad()


        return weight
    def norm_hook(self, grad):
        mean_grad = torch.mean(grad, dim=1, keepdim=True)
        return grad - mean_grad

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass

Connection.register('full', FullConnection)
Connection.register('full_connection', FullConnection)

class one_to_one_sparse(Connection):
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(one_to_one_sparse, self).__init__(pre=pre, post=post, name=name,
                                                link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                sparse_with_mask=sparse_with_mask,
                                                pre_var_name=pre_var_name, post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)
        try:
            assert self.pre_num == self.post_num
        except AssertionError:
            raise ValueError(
                'One to One connection must be defined in two groups with the same size, but the pre_num %s is not equal to the post_num %s.' % (
                self.pre_num, self.post_num))
        weight = kwargs.get('weight', None)
        self.w_mean = kwargs.get('w_mean', 0.05)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        self.is_parameter = kwargs.get('is_parameter', False)
        self.is_sparse = kwargs.get('is_sparse', True)

        if weight is None:
            # Connection weight
            self.weight = self.w_mean * np.eye(*self.shape)
        else:
            if isinstance(weight, BaseInitializer):
                raise ValueError('Sparse implementation of one_to_one_sparse connection do not support the %s '
                                 'initialization method. You can pass a numpy matrix whose diagonal is not zero.',
                                 self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight
        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value


    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass

Connection.register('one_to_one_sparse', one_to_one_sparse)


class one_to_one_mask(Connection):
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=True, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(one_to_one_mask, self).__init__(pre=pre, post=post, name=name,
                                              link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                              sparse_with_mask=sparse_with_mask, pre_var_name=pre_var_name,
                                              post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)
        try:
            assert self.pre_num == self.post_num
        except AssertionError:
            raise ValueError('One to One connection must be defined in two groups with the same size, but '
                             'the pre_num %s is not equal to the post_num %s.' % (self.pre_num, self.post_num))
        weight = kwargs.get('weight', None)
        self.w_mean = kwargs.get('w_mean', 0.05)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        # self._variables = dict()

        if weight is None:
            # Connection weight
            self.weight = self.w_mean * np.eye(*self.shape)
        else:
            if isinstance(weight, BaseInitializer):
                raise ValueError('Dense implementation of one_to_one_mask connection do not support the %s '
                                 'initialization method. You can pass a numpy matrix whose diagonal is not zero.', self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('one_to_one', one_to_one_mask)

class conv_connect(Connection):

    '''
    do the convolution connection.

    Args:
        pre(Assembly): the assembly which needs to be connected
        post(Assembly): the assembly which needs to connect the pre
        link_type(str): Conv
    Methods:
        unit_connect: define the basic connection information and add them to the connection_information.
        condition_check: check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

    '''

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['conv'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(conv_connect, self).__init__(pre=pre, post=post, name=name,
                                           link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                           sparse_with_mask=sparse_with_mask,
                                           pre_var_name=pre_var_name, post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)
        self.out_channels = kwargs.get('out_channels', None)
        self.in_channels = kwargs.get('in_channels', None)
        self.kernel_size = kwargs.get('kernel_size', [3, 3])
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.05)
        weight = kwargs.get('weight', None)

        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        self.mask = kwargs.get('mask', None)
        self.stride = kwargs.get('stride', 1)
        self.padding = kwargs.get('padding', 0)
        self.dilation = kwargs.get('dilation', 1)
        self.groups = kwargs.get('groups', 1)

        if isinstance(self.kernel_size, int):
            self.kernel_size = [self.kernel_size] * 2
        if isinstance(self.stride, int):
            self.stride = [self.stride] * 2
        if isinstance(self.padding, int):
            self.padding = [self.padding] * 2
        # self._variables = dict()
        self._constant_variables['stride[pre]'] = self.stride
        self._constant_variables['padding[pre]'] = self.padding
        self._constant_variables['dilation[pre]'] = self.dilation
        self._constant_variables['groups[pre]'] = self.groups

        if self.in_channels is None:
            from spaic.Neuron.Encoders import NullEncoder
            if len(self.pre.shape) == 1:
                raise ValueError('The shape %s of pre seems to be a flattened value. '
                                 'Give the correct shape with structure [channel, height, width] in the NeuronGroup '
                                 'class with the neuron_shape parameter' % self.pre.shape)
            elif isinstance(self.pre, NullEncoder):
                self.in_channels = self.pre.shape[2]
            else:
                self.in_channels = self.pre.shape[0]

        if self.out_channels is None:
            if len(self.post.shape) == 1:
                raise ValueError('The shape %s of post seems to be a flattened value.  '
                                 'Give the correct shape with structure [channel, height, width] in the NeuronGroup '
                                 'class with the neuron_shape parameter' % self.post.shape)
            else:
                self.out_channels = self.post.shape[0]

        self.shape = (self.out_channels, self.in_channels, self.kernel_size[0], self.kernel_size[1])

        if weight is None:
            # Connection weight
            self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                self.weight = np.empty(self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                # import math
                # bound = math.sqrt(1 / (self.in_channels * self.kernel_size[0] * self.kernel_size[1]))
                # self.bias_value = np.random.uniform(-bound, bound, self.out_channels)
                self.bias_value = np.empty(self.out_channels)

            else:
                assert (bias.size == self.out_channels), f"The size of the given bias {bias.shape} does not correspond" \
                                                     f" to the size of output_channels {self.out_channels} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

        Hin = self.pre.shape[-2]
        Win = self.pre.shape[-1]

        pool_Flag = False
        if 'avgpool' in self.synapse_name:
            pool_Flag = True
            self.pool_index = self.synapse_name.index('avgpool')

        if 'maxpool' in self.synapse_name:
            pool_Flag = True
            self.pool_index = self.synapse_name.index('maxpool')

        self.conv_index = self.synapse_name.index('conv')

        # if self.pool is not None:  # 池化
        if pool_Flag:
            self.pool_only = False
            self.pool_kernel_size = kwargs.get('pool_kernel_size', [2, 2])
            self.pool_stride = kwargs.get('pool_stride', 2)
            self.pool_padding = kwargs.get('pool_padding', 0)
            if isinstance(self.pool_kernel_size, int):
                self.pool_kernel_size = [self.pool_kernel_size] * 2
            if isinstance(self.pool_stride, int):
                self.pool_stride = [self.pool_stride] * 2
            if isinstance(self.pool_padding, int):
                self.pool_padding = [self.pool_padding] * 2

            self._constant_variables['pool_kernel_size[pre]'] = self.pool_kernel_size
            self._constant_variables['pool_stride[pre]'] = self.pool_stride
            self._constant_variables['pool_padding[pre]'] = self.pool_padding

            if self.pool_index < self.conv_index:
                # pooling before conv
                self.pool_before = True
                Ho = int((Hin + 2 * self.pool_padding[0] - self.pool_kernel_size[0]) / self.pool_stride[0] + 1)
                Wo = int((Win + 2 * self.pool_padding[1] - self.pool_kernel_size[1]) / self.pool_stride[1] + 1)
                Ho = int((Ho + 2 * self.padding[0] - self.kernel_size[0]) / self.stride[0] + 1)
                Wo = int((Wo + 2 * self.padding[1] - self.kernel_size[1]) / self.stride[1] + 1)
            else:
                # conv before pooling
                self.pool_before = False
                Ho = int((Hin + 2 * self.padding[0] - self.kernel_size[0]) / self.stride[0] + 1)  # Ho = (Hin + 2 * padding[0] - kernel_size[0]) / stride[0] + 1
                Wo = int((Win + 2 * self.padding[1] - self.kernel_size[1]) / self.stride[1] + 1)  # Wo = (Win + 2 * padding[0] - kernel_size[1]) / stride[0] + 1
                # Ho = int(Ho / self.pool_kernel_size[0])
                # Wo = int(Wo / self.pool_kernel_size[1])
                Ho = int((Ho + 2 * self.pool_padding[0] - self.pool_kernel_size[0]) / self.pool_stride[0] + 1)  # Ho = (Hin + 2 * padding[0] - kernel_size[0]) / stride[0] + 1
                Wo = int((Wo + 2 * self.pool_padding[1] - self.pool_kernel_size[1]) / self.pool_stride[1] + 1)
        else:
            # without pooling
            Ho = int((Hin + 2 * self.padding[0] - self.kernel_size[0]) / self.stride[0] + 1)
            Wo = int((Win + 2 * self.padding[1] - self.kernel_size[1]) / self.stride[1] + 1)

        post_num = int(Ho * Wo * self.out_channels)

        if self.post.num == None:
            self.post.num = post_num
            self.post.shape = (self.out_channels, Ho, Wo)

        if self.post.num != None:
            if self.post.num != post_num:
                raise ValueError(
                    "The post_group num is not equal to the output num, cannot achieve the conv connection, "
                    "the output num is %d * %d * %d " % (self.out_channels, Ho, Wo))
            else:
                self.post.shape = (self.out_channels, Ho, Wo)

    def condition_check(self, pre_group, post_group):
        '''
        check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

        Args:
            pre_group(Groups): the neuron group which need to be connected in the pre.
            post_group(Groups): the neuron group which need to connect the pre_group in the post.

        Returns: flag

        '''

        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
Connection.register('conv', conv_connect)
Connection.register('conv_connection', conv_connect)

class pool_connect(Connection):

    '''
    do the pooling operation on spiking trains.

    Args:
        pre(Assembly): the assembly which needs to be connected
        post(Assembly): the assembly which needs to connect the pre
        link_type(str): Conv
    Methods:
        unit_connect: define the basic connection information and add them to the connection_information.
        condition_check: check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

    '''

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['maxpool'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(pool_connect, self).__init__(pre=pre, post=post, name=name,
                                           link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                           sparse_with_mask=sparse_with_mask,
                                           pre_var_name=pre_var_name, post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)

        self.pool_kernel_size = kwargs.get('pool_kernel_size', [2, 2])
        self.pool_stride = kwargs.get('pool_stride', 2)
        self.pool_padding = kwargs.get('pool_padding', 0)
        if isinstance(self.pool_kernel_size, int):
            self.pool_kernel_size = [self.pool_kernel_size] * 2
        if isinstance(self.pool_stride, int):
            self.pool_stride = [self.pool_stride] * 2
        if isinstance(self.pool_padding, int):
            self.pool_padding = [self.pool_padding] * 2

        self._constant_variables['pool_kernel_size[pre]'] = self.pool_kernel_size
        self._constant_variables['pool_stride[pre]'] = self.pool_stride
        self._constant_variables['pool_padding[pre]'] = self.pool_padding
        self.pool_only = True


    def condition_check(self, pre_group, post_group):
        '''
        check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

        Args:
            pre_group(Groups): the neuron group which need to be connected in the pre.
            post_group(Groups): the neuron group which need to connect the pre_group in the post.

        Returns: flag

        '''

        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag

Connection.register('pool', pool_connect)
Connection.register('pool_connection', pool_connect)


class sparse_connect_sparse(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(sparse_connect_sparse, self).__init__(pre=pre, post=post, name=name,
                                                    link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                    sparse_with_mask=sparse_with_mask,
                                                    pre_var_name=pre_var_name, post_var_name=post_var_name,
                                                    syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.w_std = kwargs.get('w_std', 0.05)
        self.density = kwargs.get('density', 0.1)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', False)
        self.is_sparse = kwargs.get('is_sparse', True)
        # self._variables = dict()

        if weight is None:
            # Connection weight
            sparse_matrix = self.w_std * sp.rand(*self.shape, density=self.density, format='csr')
            self.weight = sparse_matrix.toarray()
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                if self.w_init == 'sparse':
                    self.weight = np.empty(self.shape)
                else:
                    raise ValueError('Sparse implementation of sparse_connection_sparse connection do not support '
                                     'the %s initialization method', self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('sparse_sparse', sparse_connect_sparse)
Connection.register('sparse_connection_sparse', sparse_connect_sparse)


class sparse_connect_mask(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(sparse_connect_mask, self).__init__(pre=pre, post=post, name=name,
                                                  link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                  sparse_with_mask=sparse_with_mask,
                                                  pre_var_name=pre_var_name, post_var_name=post_var_name,
                                                  syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.005)
        self.density = kwargs.get('density', 0.1)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        # self._variables = dict()

        if weight is None:
            # Connection weight
            # sparse_matrix = self.w_std * sp.rand(*self.shape, density=self.density, format='csr')
            # self.weight = sparse_matrix.toarray()
            # self.weight[self.weight.nonzero()] = self.weight[self.weight.nonzero()] + self.w_mean
            self.weight = self.w_std * np.random.rand(*self.shape) + self.w_mean
            sparse_mask = np.less(np.random.rand(*self.shape), self.density)
            self.weight = np.multiply(self.weight, sparse_mask)
        else:
            if isinstance(weight, BaseInitializer):
                raise ValueError('Dense implementation of sparse_connection_mask do not support the %s initialization'
                                 ' method', self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('sparse', sparse_connect_mask)
Connection.register('sparse_connection', sparse_connect_mask)


class random_connect_sparse(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(random_connect_sparse, self).__init__(pre=pre, post=post, name=name,
                                                    link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                    sparse_with_mask=sparse_with_mask,
                                                    pre_var_name=pre_var_name, post_var_name=post_var_name,
                                                    syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.probability = kwargs.get('probability', 0.1)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', False)
        self.is_sparse = kwargs.get('is_sparse', True)
        # self._variables = dict()

        if weight is None:
            # Link_parameters
            prob_weight = np.random.rand(*self.shape)
            diag_index = np.arange(min([self.pre_num, self.post_num]))
            prob_weight[diag_index, diag_index] = 1
            index = (prob_weight < self.probability)
            # Connection weight
            self.weight = np.zeros(self.shape)
            self.weight[index] = prob_weight[index]
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                if self.w_init == 'sparse':
                    self.weight = np.empty(self.shape)
                else:
                    raise ValueError('Sparse implementation of random_connect_sparse connection do not support '
                                     'the %s initialization method', self.w_init)
                self.weight = np.random.randn(*self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('random_connection_sparse', random_connect_sparse)



class random_connect_mask(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(random_connect_mask, self).__init__(pre=pre, post=post, name=name,
                                                  link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                  sparse_with_mask=sparse_with_mask,
                                                  pre_var_name=pre_var_name, post_var_name=post_var_name,
                                                  syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.probability = kwargs.get('probability', 0.1)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        # self._variables = dict()

        if weight is None:
            # Link_parameters
            prob_weight = np.random.rand(*self.shape)
            diag_index = np.arange(min([self.pre_num, self.post_num]))
            prob_weight[diag_index, diag_index] = 1
            index = (prob_weight < self.probability)
            # Connection weight
            self.weight = np.zeros(self.shape)
            self.weight[index] = prob_weight[index]
        else:
            if isinstance(weight, BaseInitializer):
                raise ValueError(
                    'Dense implementation of random_connect_mask do not support the %s initialization method',
                    self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('random', random_connect_mask)
Connection.register('random_connection', random_connect_mask)


class NullConnection(Connection):

    '''
    each neuron in the first layer is connected to each neuron in the second layer.

    Args:
        pre(Assembly): The assembly which needs to be connected
        post(Assembly): The assembly which needs to connect the pre
        link_type(str): full

    Methods:
        unit_connect: define the basic connection information and add them to the connection_information.
        condition_check: check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

    '''

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(NullConnection, self).__init__(pre=pre, post=post, name=name,
                                             link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                             sparse_with_mask=sparse_with_mask,
                                             pre_var_name=pre_var_name, post_var_name=post_var_name,
                                             syn_kwargs=syn_kwargs, **kwargs)

        pass

Connection.register('null', NullConnection)

class Spike_to_Mass(Connection):
    '''
    Connect pre of spiking neurons to post of Neural Mass model
    '''
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(Spike_to_Mass, self).__init__(pre=pre, post=post, name=name,
                                            link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                            sparse_with_mask=sparse_with_mask,
                                            pre_var_name=pre_var_name, post_var_name=post_var_name,
                                            syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        bias = kwargs.get('bias', None)
        self.tau_syn = kwargs.get('tau_syn', 6.0)
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.005)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.shape = (post.num, 1)
        self.synapse_class = []


        if weight is None:
            # Connection weight
            self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                self.weight = np.empty(self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight
        self._variables['beta[link]'] = np.exp(-0.1/self.tau_syn)
        self._variables['rate[link]'] = 0.0
        self._operations.append(['rate[link]', self.spike_to_rate, [pre_var_name+'[pre]', 'rate[link]', 'beta[link]']])
        self._operations.append([self.post_var_name+'[post]', 'mat_mult_weight', 'rate[link][updated]', 'weight[link]'])


    def spike_to_rate(self, spikes, last_rate, beta):
        rate = beta*last_rate + torch.mean(spikes, dim=1, keepdim=True)
        return rate

Connection.register('spike2mass', Spike_to_Mass)

class Mass_to_Spike(Connection):
    '''
    Connect pre of spiking neurons to post of Neural Mass model
    '''
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(Mass_to_Spike, self).__init__(pre=pre, post=post, name=name,
                                            link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                            sparse_with_mask=sparse_with_mask,
                                            pre_var_name=pre_var_name, post_var_name=post_var_name,
                                            syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        bias = kwargs.get('bias', None)
        self.tau_syn = kwargs.get('tau_syn', 6.0)
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.005)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.shape = (post.num, pre.num)
        self.synapse_class = []
        self.unit_conversion = kwargs.get('unit_conversion', 1.0)


        if weight is None:
            # Connection weight
            self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                self.weight = np.empty(self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight
        self._variables['beta[link]'] = np.exp(-0.1/self.tau_syn)
        self._variables['rate[link]'] = 0.0

        self._operations.append([post_var_name+'[post]', self.rate_to_Isyn, [pre_var_name+'[pre]', 'weight[link]']])



    def rate_to_Isyn(self, rate, weight):
        rate = rate.view(rate.shape[0], 1, self.pre.num).expand(rate.shape[0],
                                                                self.post.num, self.pre.num)
        spikes = torch.rand_like(rate).le(rate * self.unit_conversion * self.dt).type(self._backend.data_type)
        Isyn = torch.sum(spikes*weight, dim=-1)
        return Isyn
Connection.register('mass2spike', Mass_to_Spike)

class DistDepd_connect(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        super(DistDepd_connect, self).__init__(pre=pre, post=post, name=name,
                                               link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                               sparse_with_mask=sparse_with_mask,
                                               pre_var_name=pre_var_name, post_var_name=post_var_name,
                                               syn_kwargs=syn_kwargs, **kwargs)
        self.distance_weight_function = kwargs.get('distance_weight_function', None)
        self.zero_self = kwargs.get('zero_self', False)
        if self.distance_weight_function is None:
            self.distance_weight_function = self.default_dist_weight_function
        self.dist_function = kwargs.get('dist_function', 'euclidean')
        if self.dist_function == 'euclidean':
            self.dist_function = self.euclidean_dist_function
        elif self.dist_function == 'circular':
            self.dist_function = self.circular_dist_function

        self.dist_a = kwargs.get('dist_a', 0.2)
        self.dist_b = kwargs.get('dist_b', 0.4)
        self.w_amp = kwargs.get('w_amp', -0.1)
        self.pos_range = kwargs.get('pos_range', 1.0)
        self.p = kwargs.get('p', 1.0)

        # building the weight
        pre_num = pre.num
        post_num = post.num
        assert len(pre.position) > 0
        assert len(post.position) > 0

        post_pos = np.expand_dims(post.position, axis=1)
        pre_pos = np.expand_dims(pre.position, axis=0)
        weight = self.distance_weight_function(self.dist_function(pre_pos, post_pos))
        if self.p < 1.0:
            rand_mask = torch.rand_like(weight).lt(self.p)
            weight = weight*rand_mask
        self._variables['weight[link]'] = weight





    def unit_connect(self, pre_group, post_group):
        # The number of neurons in neuron group
        pre_num = pre_group.num
        post_num = post_group.num
        assert len(pre_group.position) > 0
        assert len(post_group.position) > 0
        link_num = pre_num * post_num
        shape = (post_num, pre_num)
        weight = np.zeros(shape)
        post_pos = np.expand_dims(post_group.position,  axis=1)
        pre_pos = np.expand_dims(pre_group.position, axis=0)
        weight = self.distance_weight_function(self.dist_function(pre_pos, post_pos))
        # from matplotlib import pyplot as plt
        # plt.imshow(weight)
        # plt.show()

        # The name for backend variables
        input_name = self.get_input_name(pre_group, post_group)
        weight_name = self.get_weight_name(pre_group, post_group)
        target_name = self.get_target_name(post_group)

        # The backend variable
        var_code = (weight_name, shape, weight, True, False) # (var_name, shape, value, is_parameter, is_sparse, init)
        op_code = [target_name, 'mat_mult', input_name, weight_name]
        connection_information = (pre_group, post_group, link_num, var_code, op_code)
        self.unit_connections.append(connection_information)

    def circular_dist_function(self, pre_pos, post_pos):
        if not isinstance(pre_pos, torch.Tensor):
            pre_pos = torch.tensor(pre_pos)
        if not  isinstance(post_pos, torch.Tensor):
            post_pos = torch.tensor(post_pos)

        z = torch.maximum(pre_pos, post_pos)
        k = torch.minimum(pre_pos, post_pos)
        dist = torch.minimum(z - k, self.pos_range + k - z)
        dist = torch.norm(dist, p=2, dim=-1)
        return dist

    def euclidean_dist_function(self, pre_pos, post_pos):
        if isinstance(pre_pos, torch.Tensor) and isinstance(post_pos, torch.Tensor):
            diff = pre_pos - post_pos
        else:
            diff = torch.tensor(pre_pos-post_pos)
        dist = torch.norm(diff, p=2, dim=-1)
        return dist

    def default_dist_weight_function(self, dist):
        weights = self.w_amp * (torch.exp(-dist / self.dist_a)/self.dist_a - 0.5*torch.exp(-dist / self.dist_b)/self.dist_b)
        if self.zero_self:
            weights = weights * (dist!=0).type(self._backend.data_type)
        # import matplotlib.pyplot as plt
        # plt.imshow(weights, aspect='auto')
        # plt.show()
        return weights
Connection.register('dist_depd', DistDepd_connect)

# class reconnect(Connection):
#     def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connect', 'conv', '...'), policies=[],
#                  max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
#         super(reconnect, self).__init__(pre=pre, post=post, name=name, link_type=link_type,
#                                              policies=policies, max_delay=max_delay, sparse_with_mask=sparse_with_mask, pre_var_name=pre_var_name, post_var_name=post_var_name,syn_kwargs=syn_kwargs, **kwargs)
#     def unit_connect(self, pre_group, post_group):
#         pass
#
#     def condition_check(self, pre_group, post_group):
#         pass